/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.*;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Vector;
import java.text.DateFormat;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.primitives.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.html.*;
import com.dickimawbooks.texparserlib.aux.*;

public class Abstract2Html extends L2HConverter
{
   protected Abstract2Html(TeXApp app,
     String htmlTitle, String htmlAuthor, int jmlrVolume,
     int startPage, int endPage, int year)
   {
      super(app, true, true);
      this.title     = htmlTitle;
      this.authors   = htmlAuthor;
      this.volume    = jmlrVolume;
      this.startPage = startPage;
      this.endPage   = endPage;
      this.year      = year;
   }

   protected void addPredefined()
   {
      super.addPredefined();

      putControlSequence(new AbstractEnv());

      putControlSequence(new Editor("editor"));
      putControlSequence(new Editor("editors"));
      putControlSequence(new MakeTitle());
      putControlSequence(new AtSecondOfTwo("hyperlink"));// suppress hyperlinks
   }

   public void setInputEncoding(String enc)
   {
      super.setInputEncoding(enc);

      if (enc != null && !enc.equals("utf8"))
      {
         ((BookTeXApp)getTeXApp()).warn(getParser().getLineNumber(),
            MakeJmlrBookGUI.getLabelWithValue("warning.non_utf8", enc));

      }
   }


   protected void addIgnoreable(String name, int numArgs)
   {
      addIgnoreable(name, numArgs, false);
   }

   protected void addIgnoreable(String name, int numArgs, boolean hasOpt)
   {
      putControlSequence(new A2HIgnoreable(name, numArgs, hasOpt));
   }

   protected void addIgnoreable(String name, int numArgs, boolean hasOpt,
     boolean hasStar)
   {
      putControlSequence(new A2HIgnoreable(name, numArgs, hasOpt, hasStar));
   }

   public void par() throws IOException
   {
      if (isInDocEnv())
      {
         write("<p>");
      }
   }

   public void warn(int lineNum, String msg)
   {
      ((BookTeXApp)getTeXApp()).warn(lineNum, msg);
   }

   public void unknown(int lineNum, String name) throws IOException
   {
      if (isInDocEnv())
      {
         warn(lineNum, 
            MakeJmlrBookGUI.getLabelWithValue("error.undefined", "\\"+name));
      }
      else
      {
         MakeJmlrBookGUI.debug(MakeJmlrBookGUI.getLabelWithValues(
            "error.line", ""+lineNum, 
            MakeJmlrBookGUI.getLabelWithValue("error.undefined", "\\"+name)));
      }
   }

   public static void convert(File texFile, File htmlFile, BookTeXApp app,
     String htmlTitle, String htmlAuthor, int jmlrVolume,
     int startPage, int endPage, int year)
     throws IOException
   {
      
      String base = texFile.getName();

      int idx = base.lastIndexOf(".");

      if (idx > -1)
      {
         base = base.substring(0, idx);
      }


      Abstract2Html listener = new Abstract2Html(app, htmlTitle,
       htmlAuthor, jmlrVolume, startPage, endPage, year);
      PrintWriter writer = null;

      listener.base = base;

      try
      {
         writer = new PrintWriter(htmlFile);
         listener.setWriter(writer);
         listener.setParser(new TeXParser(listener));
         listener.setParseAuxEnabled(true);

         listener.parser.parse(texFile);
      }
      finally
      {
         listener.setWriter(null);

         if (writer != null)
         {
            writer.close();
         }
      }
   }

   public ControlSequence createUndefinedCs(String name)
   {
      return new AbsUndefined(name);
   }

   public void environment(Environment env)
      throws IOException
   {
      try
      {
         env.process(parser);
      }
      finally
      {
         if (env.getName().equals("abstract"))
         {
            writeln("<p><font color=\"gray\">[abs]</font>");
            writeln("[<a href=\""+base+".pdf\">pdf</a>]");

            endDocument();
         }
      }
   }

   public void writeCodePoint(int codePoint)
     throws IOException
   {
      if (isInDocEnv())
      {
         super.writeCodePoint(codePoint);
      }
   }

   public void write(char chr)
     throws IOException
   {
      if (isInDocEnv())
      {
         super.write(chr);
      }
   }

   public void write(String str)
     throws IOException
   {
      if (isInDocEnv())
      {
         super.write(str);
      }
   }

   public void writeln(String str)
     throws IOException
   {
      if (isInDocEnv())
      {
         super.writeln(str);
      }
   }

   public static void writeHead(Writeable writeable, String htmlTitle)
    throws IOException
   {
      writeable.writeln("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\"> ");
      writeable.writeln("<html>");
      writeable.writeln("<head>");

      writeable.write("<meta http-equiv=\"Content-Type\" ");
      writeable.writeln("content=\"text/html; charset=UTF-8\" >");

      writeable.writeln("<meta name=\"generator\" content=\"MakeJmlrBookGUI\">");

      writeable.writeln("<title>"+htmlTitle+"</title>");

      // Styles

      writeable.writeln("<!-- Stylesheet -->");
      writeable.write("<link rel=\"stylesheet\" type=\"text/css\" ");
      writeable.writeln("href=\"http://www.jmlr.org/style.css\" >");

      writeable.write("<link rel=\"alternate\" type=\"application/rss+xml\" ");
      writeable.writeln("href=\"http://www.jmlr.org/jmlr.xml\" title=\"JMLR RSS\">");

      writeable.writeln("<style>. {font-family:verdana,helvetica,sans-serif}");
      writeable.writeln("a {text-decoration:none;color:#3030a0}");
      writeable.writeln("</style>");
      writeable.writeln("<style type=\"text/css\">");
      writeable.writeln("<!--");
      writeable.writeln("#fixed {");
      writeable.writeln("    position: absolute;");
      writeable.writeln("    top: 0;");
      writeable.writeln("    left: 0;");
      writeable.writeln("    width: 8em;");
      writeable.writeln("    height: 100%;");
      writeable.writeln("}");
      writeable.writeln("body > #fixed {");
      writeable.writeln("    position: fixed;");
      writeable.writeln("}");
      writeable.writeln("#content {");
      writeable.writeln("    margin-top: 1em;");
      writeable.writeln("    margin-left: 10em;");
      writeable.writeln("    margin-right: 0.5em;");
      writeable.writeln("}");
      writeable.writeln("img.jmlr {");
      writeable.writeln("    width: 7em;");
      writeable.writeln("}");
      writeable.writeln("img.rss {");
      writeable.writeln("    width: 2em;");
      writeable.writeln("}");
      writeable.writeln("-->");
      writeable.writeln("</style>");

      // MathJax

      writeable.writeln("<!-- MathJax -->");
      writeable.writeln("<script type=\"text/x-mathjax-config\">");
      writeable.writeln("MathJax.Hub.Config({tex2jax: { inlineMath: [['$','$'],['\\\\(','\\\\)']] }});");
      writeable.writeln("</script>");

      writeable.write("<script type=\"text/javascript\" src=");
      writeable.writeln(
       "\"http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML\">");
      writeable.writeln("</script>");

      // JavaScript
      writeable.writeln("<!--");
      writeable.writeln("<script language=\"JavaScript\">");
      writeable.writeln("function GoAddress(user,machine) {");
      writeable.writeln("  document.location = 'mailto:' + user + '@' + machine; }");
      writeable.writeln("</script>");
      writeable.writeln("-->");

      writeable.writeln("</head>");

      // Start of HTML body

      writeable.writeln("<body>");

      // Side Navigation

      writeable.writeln("<div id=\"fixed\">");
      writeable.writeln("<br>");
      writeable.write("<a align=\"right\" href=\"http://www.jmlr.org/\" ");
      writeable.write("target=_top><img align=\"right\" class=\"jmlr\" ");
      writeable.write("src=\"http://www.jmlr.org/jmlr.jpg\" ");
      writeable.writeln("border=\"0\" /></a> ");
      writeable.writeln("<p><br><br></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.writeln("<a href=\"http://www.jmlr.org/\">Home Page</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.writeln("<a href=\"http://www.jmlr.org/papers\">Papers</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/author-info.html\">");
      writeable.writeln("Submissions</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/news.html\">");
      writeable.writeln("News</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/editorial-board.html\">");
      writeable.writeln("Editorial Board</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/announcements.html\">");
      writeable.writeln("Announcements</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/proceedings.html\">");
      writeable.writeln("Proceedings</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/mloss\">");
      writeable.writeln("Open Source Software</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/search-jmlr.html\">");
      writeable.writeln("Search</a></p>");
      writeable.write("<p align=\"right\"> ");
      writeable.write("<a href=\"http://www.jmlr.org/manudb\">");
      writeable.writeln("Login</a></p>");
      writeable.writeln("<p><br/><br/></p>");
      writeable.writeln("<p align=\"right\">");
      writeable.write("<a href=\"http://www.jmlr.org/jmlr.xml\">");
      writeable.write("<img src=\"http://www.jmlr.org/RSS.gif\" ");
      writeable.write("class=\"rss\" alt=\"RSS Feed\">");
      writeable.writeln("</a></p>");
      writeable.writeln("</div>");

      // Content

      writeable.writeln("<div id=\"content\">");

   }

   public static void writeEndBody(Writeable writeable)
      throws IOException
   {
      Calendar date = Calendar.getInstance();
      DateFormat df = DateFormat.getDateTimeInstance();

      writeable.write("<script src=\"");
      writeable.write(
         "//ajax.googleapis.com/ajax/libs/jquery/2.0.0/jquery.min.js");
      writeable.writeln("\"></script>");

      writeable.writeln("<p align=\"center\">");
      writeable.writeln("Page last modified on "+df.format(date.getTime())+".");
      writeable.writeln("</p>");

      writeable.writeln("<table width=\"100%\">");
      writeable.writeln("<tr>");
      writeable.write("<td align=\"left\">");
      writeable.write("<font size=\"-1\"><i>");
      writeable.write("<a href=\"javascript:GoAddress('webmaster','jmlr.org');\">webmaster");
      writeable.write("jmlr.org</a></i>");
      writeable.writeln("</font></td>");
      writeable.write("<td align=\"right\"><font size=\"-1\">");
      writeable.write("Copyright @ <a target=\"_top\" ");
      writeable.write("href=\"http://www.jmlr.org/\">JMLR</a> "
       + date.get(Calendar.YEAR));
      writeable.writeln(".  All rights reserved.</font></td>");
      writeable.writeln("</tr>");
      writeable.writeln("</table>");

      writeable.writeln("</div> <!-- content -->");
      writeable.writeln("</body>");
      writeable.writeln("</html>");
   }

   public void beginDocument()
      throws IOException
   {
      super.beginDocument();

      writeHead(getWriteable(), title);
   }

   public void endDocument()
     throws IOException
   {
      writeEndBody(getWriteable());
      super.endDocument();
   }

   public void maketitle()
    throws IOException
   {
      writeln("<h1>"+title+"</h1>");
      write("<p>");
      write("<b>"+authors+"</b>; JMLR W&amp;CP "+volume);
      write(":"+startPage+"-"+endPage+", "+year+".");

      if (editor != null)
      {
         write("<br/><b>"+editorTag+":</b> "+editor);
      }

      writeln("</p>");
   }

   public void setEditor(String tag, String name)
   {
      editorTag = tag;
      editor = name;
   }

   public String getBase()
   {
      return base;
   }

   private String title, authors, editor, editorTag, base;

   private int volume, startPage, endPage, year;
}

class A2HIgnoreable extends ControlSequence
{
   public A2HIgnoreable(String name, int numArgs)
   {
      this(name, numArgs, false, false);
   }

   public A2HIgnoreable(String name, int numArgs, boolean hasOptional)
   {
      this(name, numArgs, hasOptional, false);
   }

   public A2HIgnoreable(String name, int numArgs, boolean hasOptional,
     boolean hasStar)
   {
      super(name);
      this.numArgs = numArgs;
      this.hasOptional = hasOptional;
      this.hasStar = hasStar;
   }

   public Object clone()
   {
      return new A2HIgnoreable(getName(), numArgs, hasOptional);
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      if (hasStar)
      {
         TeXObject obj = list.peek();

         if (obj != null && (obj instanceof CharObject))
         {
            if (((CharObject)obj).getCharCode() == '*')
            {
               list.popArg(parser);
            }
         }
      }

      if (hasOptional)
      {
         list.popArg(parser, '[', ']');
      }

      for (int i = 0; i < numArgs; i++)
      {
         list.popArg(parser);
      }
   }

   public void process(TeXParser parser)
     throws IOException
   {
      if (hasStar)
      {
         TeXObject obj = parser.peekStack();

         if (obj == null)
         {
            parser.fetchNext();
            obj = parser.peekStack();
         }

         if (obj != null && (obj instanceof CharObject))
         {
            if (((CharObject)obj).getCharCode() == '*')
            {
               parser.popNextArg();
            }
         }
      }

      if (hasOptional)
      {
         parser.popNextArg('[', ']');
      }

      for (int i = 0; i < numArgs; i++)
      {
         TeXObject obj = parser.popNextArg();
      }
   }

   private int numArgs;
   private boolean hasOptional, hasStar;
}

class MakeTitle extends ControlSequence
{
   public MakeTitle()
   {
      this("maketitle");
   }

   public MakeTitle(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new MakeTitle(getName());
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      process(parser);
   }

   public void process(TeXParser parser)
     throws IOException
   {
      Abstract2Html listener = (Abstract2Html)parser.getListener();

      listener.maketitle();
   }
}

class Editor extends ControlSequence
{
   public Editor()
   {
      this("editor");
   }

   public Editor(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new Editor(getName());
   }

   public void process(TeXParser parser)
     throws IOException
   {
      TeXObject arg = parser.popNextArg();

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser);
      }

      String editor;

      if (expanded == null)
      {
         editor = arg.toString(parser);
      }
      else
      {
         editor = expanded.toString(parser);
      }

      Abstract2Html listener = (Abstract2Html)parser.getListener();

      listener.setEditor(name.substring(0,1).toUpperCase()+name.substring(1),
        editor);
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      TeXObject arg = list.popArg(parser);

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser, list);
      }

      String editor;

      if (expanded == null)
      {
         editor = arg.toString(parser);
      }
      else
      {
         editor = expanded.toString(parser);
      }

      Abstract2Html listener = (Abstract2Html)parser.getListener();

      listener.setEditor(name.substring(0,1).toUpperCase()+name.substring(1),
        editor);
   }
}

class AbstractEnv extends Declaration
{
   public AbstractEnv()
   {
      this("abstract");
   }

   public AbstractEnv(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new AbstractEnv(getName());
   }

   public TeXObjectList expandonce(TeXParser parser)
    throws IOException
   {
      return null;
   }

   public TeXObjectList expandonce(TeXParser parser, TeXObjectList stack)
    throws IOException
   {
      return null;
   }

   public TeXObjectList expandfully(TeXParser parser)
    throws IOException
   {
      return null;
   }

   public TeXObjectList expandfully(TeXParser parser, TeXObjectList stack)
    throws IOException
   {
      return null;
   }

   public void process(TeXParser parser, TeXObjectList stack)
    throws IOException
   {
      process(parser);
   }

   public void process(TeXParser parser)
    throws IOException
   {
      Writeable writeable = parser.getListener().getWriteable();

      writeable.writeln("<h2>Abstract</h2>");

   }

   public void end(TeXParser parser)
     throws IOException
   {
      Abstract2Html listener = (Abstract2Html)parser.getListener();

      listener.writeln("<p><font color=\"gray\">[abs]</font>");
      listener.writeln("[<a href=\""+listener.getBase()+".pdf\">pdf</a>]");

      listener.endDocument();
   }

   public boolean isModeSwitcher()
   {
      return false;
   }
}

class AbsUndefined extends ControlSequence
{
   public AbsUndefined(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new AbsUndefined(getName());
   }

   public void process(TeXParser parser, TeXObjectList list)
    throws IOException
   {
      process(parser);
   }

   public void process(TeXParser parser)
    throws IOException
   {
      Abstract2Html listener = (Abstract2Html)parser.getListener();
      listener.write(toString());

      ((Abstract2Html)parser.getListener()).unknown(parser.getLineNumber(), name);
   }
}

