/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.*;
import java.util.HashMap;
import java.util.Vector;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.primitives.*;
import com.dickimawbooks.texparserlib.generic.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;
import com.dickimawbooks.texparserlib.latex2latex.*;

public class ArticleParser extends NewBookParser
{
   public ArticleParser(LaTeXApp texApp, Vector<LaTeXFile> styList, File outputFile)
    throws IOException
   {
      super(texApp, outputFile.getParentFile());
      this.outputFile = outputFile;
      this.styList = styList;
      this.styCheckedList = new Vector<String>();
      this.envMappings = new HashMap<String,String>();
      this.inputFiles = new Vector<File>();
      this.preEndText = new StringBuilder();
   }

   public String getOutFileName(File inFile)
   {
      return outputFile.getName();
   }

   protected void addPredefined()
   {
      super.addPredefined();

      putControlSequence(new ShortHeadings());
      putControlSequence(new JmlrHeading());
      putControlSequence(new FirstPageNo());
      putControlSequence(new TitleCs());
      putControlSequence(new AuthorCs());
      putControlSequence(new TheoremStyle());
      putControlSequence(new APBegin());
      putControlSequence(new APEnd());
      putControlSequence(new APUsePackage());
      putControlSequence(new APDef());
      putControlSequence(new APDef("gdef", true, false));
   }

   public boolean isReplaceCmd(String name)
   {
      if (name.equals("ShortHeadings")
        ||name.equals("jmlrheading")
        ||name.equals("firstpageno")
        ||name.equals("title")
        ||name.equals("author")
        ||name.equals("theoremstyle")
        ||name.equals("jmlrwcp")
        ||name.equals("jmlrnowcp")
        ||name.equals("def")
        ||name.equals("gdef"))
      {
         return true;
      }

      return super.isReplaceCmd(name);
   }

   public void documentclass(KeyValList options, String clsName)
     throws IOException
   {
      String bg = ""+parser.getBgChar();
      String eg = ""+parser.getEgChar();
      String esc = ""+parser.getEscChar();

      ControlSequence cs = parser.getControlSequence("jmlrprehyperref");

      if (cs == null && styList != null)
      {
         write(String.format("%sdef%sjmlrprehyperref%s%c%n",
           esc, esc, bg, parser.getCommentChar()));

         for (LaTeXFile sty : styList)
         {
            if (isPreHyperrefSty(sty.getName()))
            {
               usepackage(sty.getOptions(), sty.getName());

               if (isStyLoaded(sty.getName()))
               {
                  writeln();
               }
            }
            else
            {
               styCheckedList.remove(sty.getName());
            }
         }

         writeln(eg);
      }

      if (!clsName.equals("jmlr"))
      {
         String original = esc +"documentclass";
         String replacement = original;

         if (options == null)
         {
            options = new KeyValList();
         }
         else
         {
            original += "["+options.toString(parser)+"]";
            options.clear();
         }

         options.put("wcp", null);

         original += bg + clsName+eg;
         replacement += "[wcp]"+bg+"jmlr"+eg;

         substituting(original, replacement);

         super.documentclass(options, "jmlr");
      }
      else
      {
         if (options == null)
         {
            options = new KeyValList();
         }

         options.put("wcp", null);

         super.documentclass(options, clsName);
      }

      putControlSequence(new GenericCommand("jmlrwcp"));
      putControlSequence(new GenericCommand("jmlrnowcp"));

      for (File file : inputFiles)
      {
         write(String.format("%n%c %s%n",
           parser.getCommentChar(),
           MakeJmlrBookGUI.getLabelWithValue("comment.input_pseudo_sty",
            file.getName())));

         Comment comment = createComment();
         comment.appendText(
           MakeJmlrBookGUI.getLabelWithValue("comment.end_input_pseudo_sty",
            file.getName()));

         parser.push(comment);
         parser.push(getEol());
         parser.push(createGroup(file.getName()));
         parser.push(getControlSequence("input"));
      }
   }

   public boolean isPreHyperrefSty(String name)
   {
      return !(name.equals("inputenc")
        && !name.equals("fontenc")
        && !name.equals("glossaries")
        && !name.equals("bookmark"));
   }

   public LaTeXCls getLaTeXCls(String clsName)
    throws IOException
   {
      if (clsName.equals("jmlr"))
      {
         return new JmlrCls(this)
         {
            public void addDefinitions()
            {
            }

            public void preOptions() throws IOException
            {
               LaTeXParserListener listener = getListener();
               loadPreHyperrefPackages();
               listener.usepackage(null, "hyperref");
               listener.usepackage(null, "nameref");
            }

            public void processOption(String option) throws IOException
            {
            }
         };
      }

      return super.getLaTeXCls(clsName);
   }


   public void usepackage(KeyValList options, String styName) throws IOException
   {
      if (styCheckedList.contains(styName))
      {
         return;
      }

      styCheckedList.add(styName);

      if (isStyLoaded(styName))
      {
         write(String.format("%s%susepackage%s%s%s%s %s%n",
               parser.getCommentChar(), parser.getEscChar(),
               parser.getBgChar(), styName, parser.getEgChar(),
               parser.getCommentChar(),
               MakeJmlrBookGUI.getLabel("comment.already_loaded")));
         return;
      }

      if (styName.equals("inputenc"))
      {
         if (options != null && !options.containsKey("utf8"))
         {
            write(String.format("%s%susepackage%s%s%s%s %s%n",
                  parser.getCommentChar(), parser.getEscChar(),
                  parser.getBgChar(), styName, parser.getEgChar(),
                  parser.getCommentChar(),
                  MakeJmlrBookGUI.getLabelWithValue("comment.option_conflict", "utf8")));
            return;
         }
      }

      if (styName.equals("fontenc"))
      {
         if (options != null && !options.containsKey("T1"))
         {
            write(String.format("%s%susepackage%s%s%s%s %s%n",
                  parser.getCommentChar(), parser.getEscChar(),
                  parser.getBgChar(), styName, parser.getEgChar(),
                  parser.getCommentChar(),
                  MakeJmlrBookGUI.getLabelWithValue("comment.option_conflict", "T1")));
            return;
         }
      }

      if (styName.equals("algorithmicx") && isStyLoaded("algpseudocode"))
      {
         return;
      }

      String append = null;

      if (styName.equals("algpseudocode")
       || styName.equals("algorithmicx"))
      {
         append = "\\ifdef\\switchtoALGPSEUDOCODE{\\switchtoALGPSEUDOCODE}{}";
      }
      else if (styName.equals("algorithmic"))
      {
         append="\\ifdef\\switchtoALGORITHMIC{\\switchtoALGORITHMIC}{}";
      }
      else if (styName.equals("amssymb"))
      {
         append = "\\ifdef\\switchtoAMSSYMB{\\switchtoAMSSYMB}{}";
      }
      else if (styName.equals("stmaryrd"))
      {
         append = "\\ifdef\\switchtoSTMARYRD{\\switchtoSTMARYRD}{}";
      }
      else if (styName.equals("wasysym"))
      {
         append = "\\ifdef\\switchtoWASYSYM{\\switchtoWASYSYM}{}";
      }
      else if (styName.equals("mathabx"))
      {
         append = "\\ifdef\\switchtoMATHABX{\\switchtoMATHABX}{}";
      }
      else if (styName.equals("marvosym"))
      {
         append = "\\undef\\Email\\ifdef\\switchtoMARVOSYM{\\switchtoMARVOSYM}{}";
      }
      else if (styName.equals("bbding"))
      {
         append = "\\ifdef\\switchtoBBDING{\\switchtoBBDING}{}";
      }
      else if (styName.equals("ifsym"))
      {
         append = "\\ifdef\\switchtoIFSYM{\\switchtoIFSYM}{}";
      }
      else if (styName.equals("dingbat"))
      {
         append = "\\ifdef\\switchtoDINGBAT{\\switchtoDINGBAT}{}";
      }
      else if (styName.equals("wsuipa"))
      {
         append = "\\ifdef\\switchtoWSUIPA{\\switchtoWSUIPA}{}";
      }
      else if (styName.equals("wsuipa"))
      {
         append = "\\ifdef\\switchtoWSUIPA{\\switchtoWSUIPA}{}";
      }
      else if (styName.equals("MnSymbol"))
      {
         append = "\\ifdef\\switchtoMNSYMBOL{\\switchtoMNSYMBOL}{}";
      }

      if (styName.equals("jmlr2e")
       || styName.equals("algorithm")
       || styName.equals("times")
       || styName.equals("avant")
       || styName.equals("mathptm"))
      {
         substituting(""+parser.getEscChar()+"usepackage"
           +parser.getBgChar()+styName+parser.getEgChar(), "");
      }
      else
      {
         LaTeXApp app = (LaTeXApp)getTeXApp();

         if (app.isProblemSty(styName))
         {
             app.error(new NewBookException(parser,
              MakeJmlrBookGUI.getLabelWithValue("newbook.prob_sty", styName)));
         }
         else if (!app.isIgnoreSty(styName))
         {
            PackageInfo pkgInfo = app.addPackage(styName,
              options == null ? "" : options.toString(parser));

            if (styName.equals("algpseudocode") || styName.equals("algorithmicx"))
            {
               writeln("\\undef\\For");
               writeln("\\undef\\ForEach");
               writeln("\\undef\\ForAll");
               writeln("\\undef\\While");
               writeln("\\undef\\Return");
               writeln("\\undef\\Begin");
               writeln("\\undef\\If");
               writeln("\\undef\\Else");
               writeln("\\undef\\Switch");
               writeln("\\undef\\Case");
               writeln("\\undef\\Other");
            }

            if (pkgInfo != null && !pkgInfo.isInstalled())
            {
               write(
                 String.format("%s%susepackage", parser.getCommentChar(),
                    parser.getEscChar()));

               if (options != null)
               {
                  write("["+options.toString(parser)+"]");
               }

               write(String.format("%c%s%c%c %s%n",
                  parser.getBgChar(), styName, parser.getEgChar(),
                  parser.getCommentChar(), 
                  MakeJmlrBookGUI.getLabel("comment.not_installed")));

               // Try http://ctan.org/pkg/name

               boolean found = false;

               try
               {
                  if (MakeJmlrBookGUI.ctanPkgExists(styName))
                  {
                     found = true;
                  }
               }
               catch (IOException e)
               {
                  MakeJmlrBookGUI.debug(e);
               }

               String message = MakeJmlrBookGUI.getLabelWithValue(
                  found ? "newbook.missing_sty.ctan_match" : "newbook.missing_sty", styName);

               // Does the package file exist in the article's
               // directory?

               if (!found)
               {
                  File file = new File(parser.getCurrentParentFile(),
                     styName+".sty");

                  if (file.exists())
                  {
                     // Is it a package that the author has copied
                     // into the article's directory or is it a file
                     // that just has a series of \newcommand or \def
                     // commands?

                     app.message(TeXApp.MESSAGE_INFORMATION, 
                       MakeJmlrBookGUI.getLabelWithValue(
                       "newbook.checking_for_pseudo_sty", styName));

                     if (PseudoPackageParser.isPseudoPackage(getTeXApp(), file))
                     {
                        app.message(TeXApp.MESSAGE_INFORMATION,
                           MakeJmlrBookGUI.getLabelWithValue(
                             "newbook.input_pseudo_sty", file.toString()));

                        if (getDocumentClass() == null)
                        {// In definition of \jmlrprehyperref so defer
                           inputFiles.add(file);
                        }
                        else
                        {
                           parser.push(createGroup(styName+".sty"));
                           parser.push(getControlSequence("input"));
                        }

                        return;
                     }
                     else
                     {
                        app.message(TeXApp.MESSAGE_INFORMATION,
                           MakeJmlrBookGUI.getLabelWithValue(
                          "newbook.noinput_pseudo_sty", styName));
                        message = MakeJmlrBookGUI.getLabelWithValue(
                          "newbook.ignoring_sty", styName);
                     }
                  }
               }

               app.error(new NewBookException(parser, message));
            }
            else
            {
               super.usepackage(options, styName);
            }

            if (append != null)
            {
               write(append);
            }

            if (styName.equals("mdwlist"))
            {
               writeln("\\csletcs{mdwenumerate*}{enumerate*}");
               writeln("\\csletcs{endmdwenumerate*}{endenumerate*}");
               writeln("\\csundef{enumerate*}");
               writeln("\\csundef{endenumerate*}");
               addEnvMap("enumerate*", "mdwenumerate*");
            }
         }
         else
         {
            LaTeXSty sty = getLaTeXSty(styName);

            if (sty != null)
            {
               sty.load(options);
            }
         }
      }
   }

   public void shortHeadings(String title, String authors)
   {
      runningHead = title;
   }

   public void title(TeXParser parser, TeXObject opt, TeXObject fullTitle)
     throws IOException
   {
      String esc = ""+parser.getEscChar();
      write(esc+"title");

      if (opt != null || runningHead != null)
      {
         write("[");

         if (opt == null)
         {
            write(runningHead);
         }
         else
         {
            opt.process(parser);
         }

         write("]");
      }

      fullTitle.process(parser);
   }

   public void addEnvMap(String original, String replacement)
   {
      envMappings.put(original, replacement);
   }

   public String getEnvMap(String name)
   {
      return envMappings.get(name);
   }

   public void beginDocument()
     throws IOException
   {
      if (!isStyLoaded("bookmark"))
      {
         writeln("\\ifprint{}{\\usepackage{bookmark}}"
         +getParser().getCommentChar()+" "
         +MakeJmlrBookGUI.getLabel("newbook.comment.load_last"));
      }

      super.beginDocument();
   }

   public void endDocument()
     throws IOException
   {
      if (preEndText.length() > 0)
      {
         write(preEndText.toString());
      }

      super.endDocument();
   }

   public void addPreEndText(String text)
   {
      preEndText.append(text);
   }

   private String runningHead;
   private Vector<LaTeXFile> styList;
   private Vector<String> styCheckedList;

   private Vector<File> inputFiles;

   private HashMap<String,String> envMappings;

   private StringBuilder preEndText;

   private File outputFile;
}

class ShortHeadings extends L2LControlSequence
{
   public ShortHeadings()
   {
      super("ShortHeadings");
   }

   public Object clone()
   {
      return new ShortHeadings();
   }

   public void process(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      ((ArticleParser)parser.getListener()).shortHeadings(
        stack.popArg(parser).toString(parser),
        stack.popArg(parser).toString(parser));
   }

   public void process(TeXParser parser)
     throws IOException
   {
      ((ArticleParser)parser.getListener()).shortHeadings(
        parser.popNextArg().toString(parser),
        parser.popNextArg().toString(parser));
   }

}

class JmlrHeading extends L2LControlSequence
{
   public JmlrHeading()
   {
      super("jmlrheading");
   }

   public Object clone()
   {
      return new JmlrHeading();
   }

   private void process(TeXParser parser, TeXObject volume, TeXObject year,
    TeXObject pages, TeXObject submitted, TeXObject published,
    TeXObject names)
    throws IOException
   {
      String esc = ""+parser.getEscChar();
      String bg = ""+parser.getBgChar();
      String eg = ""+parser.getEgChar();

      String volumeStr = bg+volume.toString(parser)+eg;
      String yearStr = bg+year.toString(parser)+eg;
      String submittedStr = bg+submitted.toString(parser)+eg;
      String publishedStr = bg+published.toString(parser)+eg;

      String replacement = esc+"jmlrvolume"+volumeStr
        + esc+"jmlryear"+yearStr
        + esc+"jmlrsubmitted"+submittedStr
        + esc+"jmlrpublished"+publishedStr;

      parser.getListener().getWriteable().write(replacement);

      parser.getListener().getTeXApp().substituting(
        parser, toString(parser)+volumeStr+yearStr
       + bg+pages.toString(parser)+eg + submittedStr + publishedStr
       + bg+names.toString(parser)+eg,
       replacement);
   }

   public void process(TeXParser parser, TeXObjectList stack)
    throws IOException
   {
      process(parser, stack.popArg(parser), stack.popArg(parser), stack.popArg(parser),
        stack.popArg(parser), stack.popArg(parser), stack.popArg(parser));
   }

   public void process(TeXParser parser)
    throws IOException
   {
      process(parser, parser.popNextArg(), parser.popNextArg(),
        parser.popNextArg(), parser.popNextArg(),
        parser.popNextArg(), parser.popNextArg());
   }
}

class FirstPageNo extends L2LControlSequence
{
   public FirstPageNo()
   {
      super("firstpageno");
   }

   public Object clone()
   {
      return new FirstPageNo();
   }

   public void process(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      TeXObject arg = stack.popArg(parser);

      parser.getListener().getTeXApp().substituting(parser,
        toString(parser)+parser.getBgChar()+arg.toString(parser)
        +parser.getEgChar(), "");
   }

   public void process(TeXParser parser)
     throws IOException
   {
      TeXObject arg = parser.popNextArg();

      parser.getListener().getTeXApp().substituting(parser,
        toString(parser)+parser.getBgChar()+arg.toString(parser)
        +parser.getEgChar(), "");
   }
}

class TitleCs extends L2LControlSequence
{
   public TitleCs()
   {
      super("title");
   }

   public Object clone()
   {
      return new TitleCs();
   }

   private void process(TeXParser parser, TeXObject opt, TeXObject title)
      throws IOException
   {
      NewBookParser listener = (NewBookParser)parser.getListener();

      TeXObjectList builder = listener.createGroup();

      if (title instanceof TeXObjectList)
      {
         TeXObjectList list = (TeXObjectList)title;

         while (list.size() > 0)
         {
            TeXObject object = list.pop();

            if (object == null)
            {
               break;
            }

            if (object instanceof ControlSequence)
            {
               ControlSequence cs = (ControlSequence)object;

               if (cs.getName().equals("\\"))
               {
                  builder.add(listener.getControlSequence("titlebreak"));
               }
               else if (cs.getName().equals("thanks"))
               {
                  builder.add(listener.getControlSequence("titletag"));

                  Group grp = listener.createGroup();

                  grp.add(cs);

                  object = list.popArg(parser);

                  grp.add(object);

                  builder.add(grp);

               }
               else
               {
                  builder.add(object);
               }
            }
            else
            {
               builder.add(object);
            }
         }
      }
      else
      {
         builder.add(title);
      }

      ((ArticleParser)parser.getListener()).title(parser, opt, 
        builder);
   }

   public void process(TeXParser parser, TeXObjectList list)
      throws IOException
   {
      TeXObject opt = list.popArg(parser, '[', ']');

      TeXObject title = list.popArg(parser);

      process(parser, opt, title);
   }

   public void process(TeXParser parser)
      throws IOException
   {
      TeXObject opt = parser.popNextArg('[', ']');

      TeXObject title = parser.popNextArg();

      process(parser, opt, title);
   }
}

class AuthorCs extends L2LControlSequence
{
   public AuthorCs()
   {
      super("author");
   }

   public Object clone()
   {
      return new AuthorCs();
   }

   private void substitute(TeXParser parser, TeXObjectList list,
      String name, TeXObjectList builder)
     throws IOException
   {
      NewBookParser listener = (NewBookParser)parser.getListener();

      builder.add(listener.getControlSequence(
         name.substring(0,1).toUpperCase()+name.substring(1)));

      Group grp = listener.createGroup();

      builder.add(grp);

      while (list.size() > 0)
      {
         TeXObject obj = list.pop();

         if (obj == null)
         {
            break;
         }

         if (obj instanceof ControlSequence)
         {
            ControlSequence cs = (ControlSequence)obj;

            if (cs.getName().equals("email")
              ||cs.getName().equals("name")
              ||cs.getName().equals("\\")
              ||cs.getName().equals("addr")
              ||cs.getName().toLowerCase().equals("and"))
            {
               list.push(cs);
               break;
            }
            else if (cs.getName().equals("thanks"))
            {
               grp.add(listener.getControlSequence("nametag"));

               Group tagGrp = listener.createGroup();
               grp.add(tagGrp);

               tagGrp.add(cs);
               tagGrp.add(list.pop());

               grp = listener.createGroup();
               builder.add(grp);
            }
            else
            {
               grp.add(obj);
            }
         }
         else
         {
            grp.add(obj);
         }
      }
   }

   private void process(TeXParser parser, TeXObject opt, TeXObject arg)
     throws IOException
   {
      NewBookParser listener = (NewBookParser)parser.getListener();

      TeXObjectList builder = new TeXObjectList();

      if (arg instanceof TeXObjectList)
      {
         TeXObjectList list = (TeXObjectList)arg;

         while (list.size() > 0)
         {
            TeXObject object = list.pop();

            if (object == null)
            {
               break;
            }

            if (object instanceof ControlSequence)
            {
               String name = ((ControlSequence)object).getName();

               if (name.equals("name")
                || name.equals("email"))
               {
                  if (list.peek() instanceof SkippedSpaces)
                  {
                     list.pop();
                  }

                  substitute(parser, list, name, builder);
               }
               else if (name.equals("thanks"))
               {
                  builder.add(listener.getControlSequence("nametag"));

                  Group grp = listener.createGroup();
                  builder.add(grp);

                  grp.add(object);
                  grp.add(list.pop());
               }
               else
               {
                  builder.add(object);
               }
            }
            else
            {
               builder.add(object);
            }
         }
      }
      else
      {
         builder.add(arg);
      }

      Writeable writeable = parser.getListener().getWriteable();

      writeable.write(toString(parser));

      if (opt != null)
      {
         writeable.write("[");
         opt.process(parser);
         writeable.write("]");
      }


      writeable.write(parser.getBgChar());
      builder.process(parser);
      writeable.write(parser.getEgChar());
   }

   public void process(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      process(parser, stack.popArg(parser, '[', ']'), stack.popArg(parser));
   }

   public void process(TeXParser parser)
     throws IOException
   {
      process(parser, parser.popNextArg('[', ']'), parser.popNextArg());
   }
}

class TheoremStyle extends L2LControlSequence
{
   public TheoremStyle()
   {
      super("theoremstyle");
   }

   public Object clone()
   {
      return new TheoremStyle();
   }

   private void processArg(TeXParser parser, TeXObject arg)
     throws IOException
   {
      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser);
      }

      String style;

      if (expanded == null)
      {
         style = arg.toString(parser);
      }
      else
      {
         style = expanded.toString(parser);
      }

      String esc = ""+parser.getEscChar();
      String bg = ""+parser.getBgChar();
      String eg = ""+parser.getEgChar();

      StringBuilder builder = new StringBuilder();

      if (style.equals("plain"))
      {
         builder.append(esc+"theorembodyfont"+bg+"\\itshape"+eg);
         builder.append(esc+"theoremheaderfont"+bg+"\\bfseries"+eg);
         builder.append(esc+"theoremsep"+bg+""+eg);
         builder.append(esc+"theorempostheader"+bg+""+eg);
      }
      else if (style.equals("definition"))
      {
         builder.append(esc+"theorembodyfont"+bg+"\\upshape"+eg);
         builder.append(esc+"theoremheaderfont"+bg+"\\bfseries"+eg);
         builder.append(esc+"theoremsep"+bg+""+eg);
         builder.append(esc+"theorempostheader"+bg+""+eg);
      }
      else if (style.equals("remark"))
      {
         builder.append(esc+"theorembodyfont"+bg+"\\upshape"+eg);
         builder.append(esc+"theoremheaderfont"+bg+"\\itshape"+eg);
         builder.append(esc+"theoremsep"+bg+""+eg);
         builder.append(esc+"theorempostheader"+bg+""+eg);
      }

      String replacement = builder.toString();

      parser.getListener().getTeXApp().substituting(parser,
        toString(parser)+bg+style+eg,
        replacement);

      parser.getListener().getWriteable().write(replacement);
   }

   public void process(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      processArg(parser, stack.popArg(parser));
   }

   public void process(TeXParser parser)
     throws IOException
   {
      processArg(parser, parser.popNextArg());
   }
}

class APBegin extends L2LBegin
{
   public APBegin()
   {
      this("begin");
   }

   public APBegin(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new APBegin(getName());
   }

   protected void doBegin(TeXParser parser, TeXObjectList stack, String name)
     throws IOException
   {
      ArticleParser listener = (ArticleParser)parser.getListener();
      String repl = listener.getEnvMap(name);

      if (repl == null)
      {
         super.doBegin(parser, stack, name);
      }
      else
      {
         super.doBegin(parser, stack, repl);
      }
   }
}

class APEnd extends L2LEnd
{
   public APEnd()
   {
      this("end");
   }

   public APEnd(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new APEnd(getName());
   }

   protected void doEnd(TeXParser parser, String name)
     throws IOException
   {
      ArticleParser listener = (ArticleParser)parser.getListener();
      String repl = listener.getEnvMap(name);

      if (repl == null)
      {
         super.doEnd(parser, name);
      }
      else
      {
         super.doEnd(parser, repl);
      }
   }

}

class APDef extends Def
{
   public APDef()
   {
      this("def");
   }

   public APDef(String name)
   {
      super(name, true, true);
   }

   public APDef(String name, boolean isShort, boolean isLocal)
   {
      super(name, isShort, isLocal);
   }

   public Object clone()
   {
      return new APDef(getName(), isShort(), isLocal());
   }

   public void process(TeXParser parser)
      throws IOException
   {
      process(parser, parser);
   }

   public void process(TeXParser parser, TeXObjectList stack)
      throws IOException
   {
      ArticleParser listener = (ArticleParser)parser.getListener();

      TeXObject cs = stack.popStack(parser);

      if (cs == null)
      {
         stack = parser;
         cs = stack.popStack(parser);
      }

      TeXObjectList syntax = new TeXObjectList();
      TeXObject nextObject = stack.popStack(parser);

      while (!(nextObject instanceof Group))
      {
         syntax.add(nextObject);
         nextObject = stack.popStack(parser);
      }

      Group group = (Group)nextObject;

      TeXObjectList definition = group.toList();

      if (!(cs instanceof ControlSequence))
      {
         listener.getTeXApp().error(new LaTeXSyntaxException(parser,
           LaTeXSyntaxException.ERROR_UNACCESSIBLE,
           cs.toString()));

         listener.write(toString(parser));
         listener.write(cs.toString(parser));
         listener.write(syntax.toString(parser));
         listener.write(group.toString(parser));
         return;
      }

      ControlSequence cmd = (ControlSequence)cs;

      if (cmd.getName().equals("jmlrprehyperref"))
      {
         parser.putControlSequence(isLocal(),
           new GenericCommand(((ControlSequence)cs).getName(),
             syntax, definition));

         listener.write(toString(parser));
         listener.write(cs.toString(parser));

         if (stack == parser)
         {
            group.process(parser);
         }
         else
         {
            group.process(parser, stack);
         }

         return;
      }

      String message = null;

      if (Accent.isAccentCommand(cmd.getName()))
      {
         message = MakeJmlrBookGUI.getLabelWithValue(
            isLocal() ? "warning.accent_redef" : "warning.accent_gdef",
            cmd.getName());
      }
      else if (!isLocal())
      {
         message = MakeJmlrBookGUI.getLabelWithValue(
            "warning.gdef", cmd.getName());
      }

      if (!isLocal())
      {
         listener.write(String.format("%clet%cJMLRORG%s%c%s%n",
           parser.getEscChar(), parser.getEscChar(),
           cmd.getName(), parser.getEscChar(), cmd.getName()));

         listener.addPreEndText(String.format("%cglobal%clet%c%s%cJMLRORG%s%n",
           parser.getEscChar(), parser.getEscChar(), parser.getEscChar(),
           cmd.getName(), parser.getEscChar(), cmd.getName()));
      }

      if (message != null)
      {
         listener.getTeXApp().warning(parser, message);
         listener.write(String.format("%c %s%n", parser.getCommentChar(), 
            message));
      }

      listener.write(String.format("%c%s%c%s%s%s",
        parser.getEscChar(), getName(), 
        parser.getEscChar(), cmd.getName(),
        syntax.toString(parser), group.toString(parser)));
   }
}

class APUsePackage extends UsePackage
{
   public APUsePackage()
   {
      super();
   }

   public APUsePackage(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new APUsePackage(getName());
   }

   public void preProcess(TeXParser parser, TeXObjectList stack,
      KeyValList keyvalList, String styNameList)
   throws IOException
   {
      // Move any following comment before \\usepackage
      TeXObject object = null;

      if (parser == stack)
      {
         if (parser.size() == 0)
         {
            parser.fetchNext();
         }
      }

      object = stack.firstElement();

      if (object instanceof Comment)
      {
         object = stack.pop();
         object.process(parser);
      }
   }

}
