/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.util.*;
import java.io.*;
import java.net.*;
import javax.swing.tree.*;

public class BookRoot extends FileElement implements Iterable<File>
{
   public BookRoot(MakeJmlrBookGUI application)
   {
      super(application);
      init("Untitled", null);
   }

   public BookRoot(MakeJmlrBookGUI application, File dir)
     throws IOException
   {
      super(application);
      init(dir.getName(), dir);
   }

   private void init(String name, File dir)
   {
      setName(name);
      location = dir;

      book = null;

      children = new Vector<BookFile>();
   }

   public Book setMainFile(BookFile bookFile)
     throws IOException
   {
      book = bookFile.toBook();

      return book;
   }

   public File getFile()
   {
      return location == null ? new File(getName()) : location;
   }

   public File getLocation()
   {
      return location;
   }

   public void setLocation(File dir)
   {
      location = dir;
   }

   public TreeNode getParent()
   {
      return null;
   }

   public BookRoot getRoot()
   {
      return this;
   }

   public Book getBook()
   {
      return book;
   }

   public BookFile findMainFile()
   {
      for (BookFile child : children)
      {
         if (child.isDirectory() || !child.getName().endsWith(".tex"))
         {
            continue;
         }

         app.debug("Checking "+child.getName());

         BufferedReader in = null;

         try
         {
            in = new BufferedReader(new FileReader(child.getFile()));

            String line;

            while ((line = in.readLine()) != null)
            {
               String docTag = "\\documentclass";

               int idx = line.indexOf(docTag);

               if (idx > -1)
               {
                  if (line.substring(0, idx).contains("%"))
                  {
                     continue;
                  }

                  line = line.substring(idx + docTag.length());

                  line.trim();

                  if (line.length() == 0)
                  {
                     line = in.readLine();

                     if (line == null)
                     {
                        break;
                     }
                  }

                  if (line.contains("{jmlrbook}"))
                  {
                     app.debug("Found it!");
                     return child;
                  }
                  else
                  {
                     break;
                  }
               }
            }
         }
         catch (IOException e)
         {
            continue;
         }
         finally
         {
            try
            {
               if (in != null)
               {
                  in.close();
               }
            }
            catch (IOException e)
            {
            }
         }
      }

      return null;
   }

   public Iterator<File> iterator()
   {
      return new BookFileIterator(this);
   }

   // Make a copy of this root
   // optionally filtering it

   public BookRoot copy()
      throws IOException
   {
      return copy(null);
   }

   public BookRoot copy(java.io.FileFilter filter)
      throws IOException
   {
      BookRoot root = new BookRoot(getApplication(), getLocation());

      copy(root, this, filter);

      // remove empty elements

      root.trim();

      return root;
   }

   public static void copy(FileElement newElement, FileElement orgElement,
      java.io.FileFilter filter)
      throws IOException
   {
      if (!newElement.isDirectory())
      {
         return;
      }

      for (BookFile child : orgElement.children)
      {
         File file = child.getFile();

         if (child.isDirectory())
         {
            BookFile bookFile = new BookFile(newElement.getApplication(), file);

            newElement.addChild(bookFile);

            copy(bookFile, child, filter);
         }
         else
         {
            boolean addChild = true;

            if (filter != null)
            {
               addChild = filter.accept(file);
            }

            if (addChild)
            {
               BookFile bookFile = new BookFile(newElement.getApplication(), file);

               newElement.addChild(bookFile);
            }
         }
      }
   }

   public TreePath getTreePath()
   {
      return new TreePath(this);
   }

   public FileElement walk(BookWalkListener listener)
   {
      return walk(this, listener);
   }

   public static FileElement walk(FileElement start, BookWalkListener listener)
   {
      if (listener.walkAction(start) != null)
      {
         // walk terminated prematurely

         return start;
      }

      if (!start.isDirectory())
      {
         // come to the end of this trail

         return null;
      }

      for (BookFile child : start.children)
      {
         FileElement element = walk(child, listener);

         if (element != null)
         {
            // walk terminated prematurely

            return element;
         }
      }

      return null;
   }

   public void sort()
   {
      walk
      (
         new BookWalkListener()
         {
            public FileElement walkAction(FileElement element)
            {
               if (element.isDirectory())
               {
                  Collections.sort(element.children);
               }

               return null;
            }
         }
      );
   }

   public void deselectAll()
   {
      walk(new BookWalkListener()
      {
         public FileElement walkAction(FileElement element)
         {
            element.setNodeSelected(false);
            return element;
         }
      });
   }

   /**
    * Base directory of JMLR book.
    */
   protected File location;

   protected Book book;
}

class BookFileIterator implements Iterator<File>
{
   public BookFileIterator(BookRoot root)
   {
      this.root = root;

      currentFile = (root.getChildCount() == 0
                  ? null : root.children.get(0));

      while (currentFile != null && currentFile.isDirectory())
      {
         descend();
      }
   }

   public boolean hasNext()
   {
      return currentFile != null;
   }

   public File next()
      throws NoSuchElementException
   {
      if (currentFile == null)
      {
         throw new NoSuchElementException();
      }

      File file = currentFile.getFile();

      shift();

      return file;
   }

   private void descend()
   {
      if (currentFile.getChildCount() == 0)
      {
         shift();
      }
      else
      {
         currentFile = currentFile.children.get(0);
      }
   }

   private void shift()
   {
      FileElement parent = (FileElement)currentFile.getParent();

      if (parent == null)
      {
         currentFile = null;
      }

      int i = parent.getIndex(currentFile)+1;

      if (i == parent.getChildCount())
      {
         if (parent instanceof BookFile)
         {
            currentFile = (BookFile)parent;

            shift();
         }
         else
         {
            currentFile = null;
         }
      } 

      while (currentFile != null && currentFile.isDirectory())
      {
         descend();
      }
   }

   public void remove()
    throws UnsupportedOperationException,IllegalStateException
   {
      throw new UnsupportedOperationException();
   }

   private BookRoot root;
   private BookFile currentFile;
}
