/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.util.*;
import java.io.*;
import java.net.*;
import java.nio.file.*;
import java.text.*;
//import java.util.concurrent.*;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

public abstract class FileElement 
   implements TreeNode,CheckBoxNode
{
   private FileElement() {}

   protected FileElement(MakeJmlrBookGUI application)
   {
      app = application;
   }

   public MakeJmlrBookGUI getApplication()
   {
      return app;
   }

   public String getName()
   {
      return name;
   }

   public void setName(String name)
   {
      this.name = name;
   }

   public void rename(String newName)
      throws IOException
   {
      Path source = getFile().toPath();

      Files.move(source, source.resolveSibling(newName));

      name = newName;
   }

   public abstract File getFile();

   public boolean isDirectory()
   {
      return children != null;
   }

   public boolean getAllowsChildren()
   {
      return isDirectory();
   }

   public boolean isLeaf()
   {
      return children == null;
   }

   public TreeNode getChildAt(int childIndex)
   {
      if (!isDirectory())
      {
         return null;
      }

      return children.get(childIndex);
   }

   public int getIndex(TreeNode child)
   {
      if (!isDirectory())
      {
         return -1;
      }

      return children.indexOf(child);
   }

   public Enumeration<BookFile> children()
   {
      return children == null ? null : children.elements();
   }

   public Vector<BookFile> childList()
   {
      return children;
   }

   public int getChildCount()
   {
      return children == null ? 0 : children.size();
   }

   public abstract TreeNode getParent();

   public FileElement getParentElement()
   {
      return (FileElement)getParent();
   }

   public abstract File getLocation();

   public BookFile addChild(String childName, boolean isDir)
      throws IOException
   {
      if (children == null)
      {
         throw new IOException("Can't add child to "
            +name+": not a directory");
      }

      BookFile child = new BookFile(app, this, childName, isDir);
      children.add(child);

      return child;
   }

   public void addChild(BookFile child)
      throws IOException
   {
      if (children == null)
      {
         throw new IOException("Can't add child to "
            +name+": not a directory");
      }

      children.add(child);
      child.parent = this;
   }

   public boolean removeChild(BookFile child)
   {
      return children == null ? false : children.remove(child);
   }

   public void addSelected(FileElement element)
     throws IOException
   {
      if (element.isNodeSelected())
      {
         boolean isDir = element.isDirectory();

         BookFile bookFile = new BookFile(app, this, element.getName(), 
            isDir);

         bookFile.setNodeSelected(true);

         addChild(bookFile);

         File file = bookFile.getFile();

         if (isDir)
         {
            file.mkdirs();

            for (FileElement child : element.children)
            {
               bookFile.addSelected(child);
            }
         }
         else
         {
            app.verboseCopyFile(element.getFile(), file);
         }
      }
   }

   public boolean hasSelectedDescendant()
   {
      if (isDirectory())
      {
         Vector<FileElement> subDirs = new Vector<FileElement>(getChildCount());

         for (FileElement child: children)
         {
            if (child.selected)
            {
               return true;
            }

            if (child.isDirectory())
            {
               subDirs.add(child);
            }
         }

         for (FileElement child : subDirs)
         {
            if (child.hasSelectedDescendant())
            {
               return true;
            }
         }
      }

      return false;
   }

   public void populate(FileFilter filter)
   {
      if (children == null)
      {
         return;
      }

      app.setInfo(app.getLabelWithValue("message.populating", getName()));

      Vector<BookFile> stack = new Vector<BookFile>();

      for (BookFile child : children)
      {
         stack.add(child);
      }

      File file = getFile();

      File[] fileList = file.listFiles(filter);

      if (fileList == null)
      {
         return;
      }

      for (int i = 0; i < fileList.length; i++)
      {
         BookFile bookFile = null;

         for (int j = stack.size()-1; j >= 0; j--)
         {
            BookFile child = stack.get(j);

            if (child.getName().equals(fileList[i].getName()))
            {
               bookFile = child;
               stack.remove(j);
               break;
            }
         }

         if (bookFile == null)
         {
            bookFile = new BookFile(app, fileList[i]);
            bookFile.parent = this;
            children.add(bookFile);
         }

         if (bookFile.isDirectory())
         {
            bookFile.populate(filter);
         }
      }
   }

   public void trim()
   {
      if (children == null)
      {
         return;
      }

      for (int i = getChildCount()-1; i >= 0; i--)
      {
         BookFile child = children.get(i);

         if (child.isDirectory())
         {
            if ( child.getChildCount() == 0)
            {
               removeChild(child);
            }
            else
            {
               child.trim();
            }
         }
      }
   }

   public void prune()
   {
      if (children == null)
      {
         return;
      }

      app.setInfo(app.getLabelWithValue("message.pruning", getName()));

      for (int i = children.size()-1; i >= 0; i--)
      {
         BookFile child = children.get(i);

         File file = child.getFile();

         if (!file.exists())
         {
            children.remove(i);
            continue;
         }

         if (child.isDirectory())
         {
            child.prune();
         }
      }
   }

   public String toString()
   {
      return name;
   }

   public String getNodeText()
   {
      return name;
   }

   public void setNodeText(String text)
   {
      name = text;
   }

   public boolean isNodeSelected()
   {
      return selected;
   }

   public void setNodeSelected(boolean isSelected)
   {
      setNodeSelected(isSelected, false);
   }

   public void setNodeSelected(boolean isSelected, boolean recurse)
   {
      selected = isSelected;

      if (recurse && isDirectory())
      {
         for (BookFile child : children)
         {
            child.setNodeSelected(isSelected, recurse);
         }
      }
   }

   public boolean equals(Object obj)
   {
      if (this == obj)
      {
         return true;
      }

      if (obj == null || !(obj instanceof FileElement))
      {
         return false;
      }

      FileElement element = (FileElement)obj;

      if (!name.equals(element.name))
      {
         return false;
      }

      return getParentElement() == element.getParentElement();
   }

   public String getInfo()
   {
      File file = getFile();

      String text = 
         "<tr><th align=\"left\">"
         + MakeJmlrBookGUI.getLabel("fileview.file")
         +"</th><td>"
         +  file.getAbsolutePath()
         + "</td></tr>";

      if (file.exists())
      {
         long size = file.length();

         if (size > 0)
         {
            int grp = (int) (Math.log10(size)/Math.log10(1024));

            text +=
              "<tr><th align=\"left\">"
               + MakeJmlrBookGUI.getLabel("fileview.size")
               + "</th><td>"
               + fileSizeFormat.format(size/Math.pow(1024, grp))
                 + " "+fileSizeUnits[grp]
               + "</td></tr>";
         }

         text += "<tr><th align=\"left\">"
            + MakeJmlrBookGUI.getLabel("fileview.last_mod") 
            + "</th><td>"
            + dateFormat.format(new Date(file.lastModified()))
            + "</td></tr>";
      }

      return "<html><table>"+text+"</table></html>";
   }

   public void view()
     throws IOException
   {
      app.viewFile(getFile().toURI().toURL());
   }

   public BookFile findFile(File file)
   {
      BookFile bookFile = null;

      for (BookFile child: children)
      {
         File childFile = child.getFile();

         if (childFile.equals(file))
         {
            return child;
         }

         if (child.isDirectory())
         {
            bookFile = child.findFile(file);

            if (bookFile != null)
            {
               return bookFile;
            }
         }
      }

      return bookFile;
   }

   public abstract BookRoot getRoot();

   public abstract TreePath getTreePath();

   protected MakeJmlrBookGUI app;

   private String name;

   protected Vector<BookFile> children;

   protected boolean selected;

   protected static DateFormat dateFormat = DateFormat.getDateTimeInstance();

   protected static DecimalFormat fileSizeFormat = new DecimalFormat("#,##0.#");

   private static final String[] fileSizeUnits
      = new String[] {"B", "KB", "MB", "GB", "TB"};
}
