/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

public class FileViewer extends JTree 
   implements MouseListener,ActionListener
{
   public FileViewer(MakeJmlrBookGUI application)
   {
      super();

      app = application;

      messageArea = new JEditorPane();
      messageArea.setContentType("text/html");
      messageArea.setEditable(false);
      messageArea.setBorder(null);
      messageArea.setOpaque(false);
      messageArea.addHyperlinkListener(app);

      setToolTipText("");
/*
      setDragEnabled(true);
      setTransferHandler(new FileElementTransferHandler(app));
      setDropMode(DropMode.ON);
*/

      addMouseListener(this);

      popupMenu = new JPopupMenu();

      propItem = createMenuItem("fileview.popup", "properties");
      popupMenu.add(propItem);

      popupMenu.addSeparator();

      renameItem = createMenuItem("fileview.popup", "rename");
      popupMenu.add(renameItem);

      fileChooser = new JFileChooser();

      updateBook();
   }

   private JMenuItem createMenuItem(String parent, String name)
   {
      JMenuItem item = new JMenuItem(app.getLabel(parent, name));
      item.setMnemonic(app.getMnemonic(parent, name));
      item.addActionListener(this);
      item.setActionCommand(name);

      String tooltip = app.getLabelWithAlt(
         parent == null ? name : parent+"."+name,
         null);

      if (tooltip != null)
      {
         item.setToolTipText(tooltip);
      }

      return item;
   }

   public void refresh()
   {
      treeDidChange();
   }

   public void mousePressed(MouseEvent evt)
   {
      if (evt.isPopupTrigger())
      {
         showPopup(evt);
      }
   }

   public void mouseReleased(MouseEvent evt)
   {
      if (evt.isPopupTrigger())
      {
         showPopup(evt);
      }
   }

   public void mouseExited(MouseEvent evt)
   {
   }

   public void mouseEntered(MouseEvent evt)
   {
   }

   public void mouseClicked(MouseEvent evt)
   {
      if (evt.getButton() == MouseEvent.BUTTON1
           && evt.getClickCount() == 2)
      {
         TreePath path = getPathForLocation(evt.getX(), evt.getY());

         if (path == null)
         {
            return;
         }

         Object obj = path.getLastPathComponent();

         if (obj == null || !(obj instanceof FileElement))
         {
            return;
         }

         FileElement element = (FileElement)obj;

         if (!element.isDirectory())
         {
            try
            {
               element.view();
            }
            catch (Exception e)
            {
               app.error(e);
            }
         }
      }
   }

   private void displayFileInfo(TreePath path)
   {
      if (path == null)
      {
         return;
      }

      Object obj = path.getLastPathComponent();

      if (obj == null || !(obj instanceof FileElement))
      {
         return;
      }

      messageArea.setText(((FileElement)obj).getInfo());

      JOptionPane.showMessageDialog(this, messageArea,
         app.getLabel("fileview.properties"), JOptionPane.PLAIN_MESSAGE);
   }

   private void showPopup(MouseEvent evt)
   {
      TreePath path = getPathForLocation(evt.getX(), evt.getY());

      propItem.setEnabled(path != null);

      popupMenu.show(this, evt.getX(), evt.getY());
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null)
      {
         return;
      }

      if (action.equals("properties"))
      {
         displayFileInfo(getSelectionPath());
      }
      else if (action.equals("rename"))
      {
         TreePath path = getSelectionPath();

         if (path == null || !(path.getLastPathComponent() instanceof FileElement))
         {
            return;
         }

         Object obj = path.getLastPathComponent();

         if (!(obj instanceof FileElement))
         {
            return;
         }

         FileElement element = (FileElement)obj;

         String newName = JOptionPane.showInputDialog(this, 
            element.isDirectory() 
              ? app.getLabel("fileview.folder.name")
              : app.getLabel("fileview.file.name"),
            element.getName());

         if (newName == null || newName.equals(element.getName()))
         {
            return;
         }

         try
         {
            element.rename(newName);
            updateBook();
            expandElement(element);
            addSelectionPath(element.getTreePath());
         }
         catch (Exception e)
         {
            app.error(e);
         }
      }
   }

   public void selectAll()
   {
      int idx = getRowCount()-1;

      if (idx < 0)
      {
         return;
      }

      setSelectionInterval(0, idx);
   }

   public void clear()
   {
      setModel(new DefaultTreeModel(new DefaultMutableTreeNode()));
      setRootVisible(false);
   }

   public void updateBook()
   {
      BookRoot root = app.getBookRoot();

      if (root == null)
      {
         clear();
      }
      else
      {
         app.debug("Sorting file list");
         root.sort();

         setModel(new DefaultTreeModel(root));
         setRootVisible(true);
         app.debug("Update completed");
      }
   }
   
   public String getToolTipText(MouseEvent evt)
   {
      TreePath path = getPathForLocation(evt.getX(), evt.getY());

      if (path == null)
      {
         return null;
      }

      Object obj = path.getLastPathComponent();

      if (obj == null || !(obj instanceof FileElement))
      {
         return null;
      }

      return ((FileElement)obj).getInfo();
   }

   public void expandElement(FileElement element)
   {
      if (element == null)
      {
         return;
      }

      TreePath treePath = element.getTreePath();

      int idx = getRowForPath(treePath);

      if (idx == -1)
      {
         expandElement(element.getParentElement());

         idx = getRowForPath(treePath);
      }

      if (idx > -1)
      {
         expandRow(idx);
      }
   }

   private MakeJmlrBookGUI app;
   private JPopupMenu popupMenu;

   private JFileChooser fileChooser;

   private JMenuItem propItem, renameItem;

   private JEditorPane messageArea;
}
